﻿// A Force-Directed Diagram Layout Algorithm
// Bradley Smith - 2010/07/01

using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Media;

/// <summary>
/// Represents a simple diagram consisting of nodes and connections, implementing a 
/// force-directed algorithm for automatically arranging the nodes.
/// </summary>
public class Diagram {

    private const double ATTRACTION_CONSTANT = 0.5;		// spring constant
    private const double REPULSION_CONSTANT = 10000;	// charge constant
    private const double DEFAULT_DAMPING = 0.5;
    private const int DEFAULT_SPRING_LENGTH = 50;
    private const int DEFAULT_MAX_ITERATIONS = 500;

    private List<Node> mNodes;

    /// <summary>
    /// Gets or sets the damping variable for the force calculations.
    /// </summary>
    public double Damping { get; set; }
    /// <summary>
    /// Gets or sets the spring length variable for the force calculations.
    /// </summary>
    public int SpringLength { get; set; }
    /// <summary>
    /// Gets or sets the maximum number of iterations the algorithm will run for before terminating.
    /// </summary>
    public int MaxIterations { get; set; }
    /// <summary>
    /// Gets or sets the attraction constant for the force calculations.
    /// </summary>
    public double Attraction { get; set; }
    /// <summary>
    /// Gets or sets the repulsion constant for the force calculations.
    /// </summary>
    public double Repulsion { get; set; }
    /// <summary>
    /// Gets or sets the scaling factor used when drawing the diagram.
    /// </summary>
    public double DrawScalingFactor {
        get;
        set;
    }
	/// <summary>
	/// Gets a read-only collection of the nodes in this Diagram.
	/// </summary>
	public IList<Node> Nodes {
		get {
			return mNodes.AsReadOnly();
		}
	}

	/// <summary>
	/// Initialises a new instance of the Diagram class.
	/// </summary>
	public Diagram() {
        Damping = DEFAULT_DAMPING;
        SpringLength = DEFAULT_SPRING_LENGTH;
        MaxIterations = DEFAULT_MAX_ITERATIONS;
        Attraction = ATTRACTION_CONSTANT;
        Repulsion = REPULSION_CONSTANT;

		mNodes = new List<Node>();
        DrawScalingFactor = 1.0;
	}

	/// <summary>
	/// Adds the specified Node to this Diagram.
	/// </summary>
	/// <param name="node">The Node to add to the diagram.</param>
	/// <returns>True if the node was added, false if the node is already on this Diagram.</returns>
	public bool AddNode(Node node) {
		if (node == null) throw new ArgumentNullException("node");
		
		if (!mNodes.Contains(node)) {
			// add node, associate with diagram, then add all connected nodes
			mNodes.Add(node);
			node.Diagram = this;
			foreach (Node child in node.Connections) AddNode(child);
			return true;
		}
		else {
			return false;
		}
	}

    /// <summary>
	/// Runs the force-directed layout algorithm on this Diagram, using the specified parameters.
	/// </summary>
	/// <param name="damping">Value between 0 and 1 that slows the motion of the nodes during layout.</param>
	/// <param name="springLength">Value in pixels representing the length of the imaginary springs that run along the connectors.</param>
	/// <param name="maxIterations">Maximum number of iterations before the algorithm terminates.</param>
	/// <param name="deterministic">Whether to use a random or deterministic layout.</param>
    public void Arrange(int? randomSeed = 0) {
		// random starting positions can be made deterministic by seeding System.Random with a constant
        Random rnd = randomSeed.HasValue ? new Random(randomSeed.Value) : new Random();

		// copy nodes into an array of metadata and randomise initial coordinates for each node
		NodeLayoutInfo[] layout = new NodeLayoutInfo[mNodes.Count];
		for (int i = 0; i < mNodes.Count; i++) {
			layout[i] = new NodeLayoutInfo(mNodes[i], new Vector(), new Point());
			layout[i].Node.Location = new Point(rnd.Next(-50, 50), rnd.Next(-50, 50));
		}

		int stopCount = 0;
		int iterations = 0;

		while (true) {
			double totalDisplacement = 0;

			for (int i=0; i<layout.Length; i++) {
				NodeLayoutInfo current = layout[i];

				// express the node's current position as a vector, relative to the origin
				Vector currentPosition = new Vector(CalcDistance(new Point(), current.Node.Location), GetBearingAngle(new Point(), current.Node.Location));
				Vector netForce = new Vector(0, 0);

				// determine repulsion between nodes
				foreach (Node other in mNodes) {
					if (other != current.Node) netForce += CalcRepulsionForce(current.Node, other);
				}

				// determine attraction caused by connections
				foreach (Node child in current.Node.Connections) {
					netForce += CalcAttractionForce(current.Node, child, SpringLength);
				}
				foreach (Node parent in mNodes) {
					if (parent.Connections.Contains(current.Node)) netForce += CalcAttractionForce(current.Node, parent, SpringLength);
				}

				// apply net force to node velocity
				current.Velocity = (current.Velocity + netForce) * Damping;

				// apply velocity to node position
				current.NextPosition = (currentPosition + current.Velocity).ToPoint();
			}

			// move nodes to resultant positions (and calculate total displacement)
			for (int i = 0; i < layout.Length; i++) {
				NodeLayoutInfo current = layout[i];

				totalDisplacement += CalcDistance(current.Node.Location, current.NextPosition);
				current.Node.Location = current.NextPosition;
			}

            // center the diagram around the origin
            Rect logicalBounds = GetDiagramBounds();
            Point midPoint = new Point(logicalBounds.X + (logicalBounds.Width / 2), logicalBounds.Y + (logicalBounds.Height / 2));

            foreach (Node node in mNodes) {
                node.Location = new Point(node.Location.X - midPoint.X, node.Location.Y - midPoint.Y);
            }

			iterations++;
			if (totalDisplacement < 10) stopCount++;
			if (stopCount > 15) break;
			if (iterations > MaxIterations) break;
		}
	}

	/// <summary>
	/// Calculates the attraction force between two connected nodes, using the specified spring length.
	/// </summary>
	/// <param name="x">The node that the force is acting on.</param>
	/// <param name="y">The node creating the force.</param>
	/// <param name="springLength">The length of the spring, in pixels.</param>
	/// <returns>A Vector representing the attraction force.</returns>
	private Vector CalcAttractionForce(Node x, Node y, double springLength) {
		int proximity = Math.Max(CalcDistance(x.Location, y.Location), 1);

		// Hooke's Law: F = -kx
		double force = Attraction * Math.Max(proximity - springLength, 0);
		double angle = GetBearingAngle(x.Location, y.Location);

		return new Vector(force, angle);
	}

	/// <summary>
	/// Calculates the distance between two points.
	/// </summary>
	/// <param name="a">The first point.</param>
	/// <param name="b">The second point.</param>
	/// <returns>The pixel distance between the two points.</returns>
	public static int CalcDistance(Point a, Point b) {
		double xDist = (a.X - b.X);
		double yDist = (a.Y - b.Y);
		return (int)Math.Sqrt(Math.Pow(xDist, 2) + Math.Pow(yDist, 2));
	}

	/// <summary>
	/// Calculates the repulsion force between any two nodes in the diagram space.
	/// </summary>
	/// <param name="x">The node that the force is acting on.</param>
	/// <param name="y">The node creating the force.</param>
	/// <returns>A Vector representing the repulsion force.</returns>
	private Vector CalcRepulsionForce(Node x, Node y) {
		int proximity = Math.Max(CalcDistance(x.Location, y.Location), 1);

		// Coulomb's Law: F = k(Qq/r^2)
		double force = -(Repulsion / Math.Pow(proximity, 2));
		double angle = GetBearingAngle(x.Location, y.Location);

		return new Vector(force, angle);
	}

	/// <summary>
	/// Removes all nodes and connections from the diagram.
	/// </summary>
	public void Clear() {
		mNodes.Clear();
	}

	/// <summary>
	/// Determines whether the diagram contains the specified node.
	/// </summary>
	/// <param name="node">The node to test.</param>
	/// <returns>True if the diagram contains the node.</returns>
	public bool ContainsNode(Node node) {
		return mNodes.Contains(node);
	}

    /// <summary>
    /// Calculates the scaling factor for the specified bounds.
    /// </summary>
    /// <param name="bounds"></param>
    /// <returns></returns>
    private double GetScalingFactor(Rect bounds) {
        Rect logicalBounds = GetDiagramBounds();
        double scale = 1;
        if (logicalBounds.Width > logicalBounds.Height) {
            if (logicalBounds.Width != 0) scale = (double)Math.Min(bounds.Width, bounds.Height) / (double)logicalBounds.Width;
        }
        else {
            if (logicalBounds.Height != 0) scale = (double)Math.Min(bounds.Width, bounds.Height) / (double)logicalBounds.Height;
        }

        return scale;
    }

    /// <summary>
    /// Automatically scales the diagram to its bounds.
    /// </summary>
    /// <param name="bounds"></param>
    /// <returns></returns>
    public double AutoScale(Rect bounds) {
        return DrawScalingFactor = GetScalingFactor(bounds);
    }

	/// <summary>
	/// Returns a Drawing object representing the diagram, centering within the specified bounds.
	/// </summary>
	/// <param name="bounds">Bounds in which to draw the diagram.</param>
    /// <param name="autoScale">Determines whether the diagram automatically scales to its bounds.</param>
	public Drawing MakeDrawing(Rect bounds, bool autoScale = true) {
        DrawingGroup drawing = new DrawingGroup();
        DrawingGroup connectors = new DrawingGroup();
        DrawingGroup nodes = new DrawingGroup();

        Point center = new Point(bounds.X + (bounds.Width / 2), bounds.Y + (bounds.Height / 2));

		// determine the scaling factor
        if (autoScale) DrawScalingFactor = GetScalingFactor(bounds);

		// draw all of the connectors first
		foreach (Node node in mNodes) {
            Point source = ScalePoint(node.Location, DrawScalingFactor);

			// connectors
			foreach (Node other in node.Connections) {
                Point destination = ScalePoint(other.Location, DrawScalingFactor);
                connectors.Children.Add(node.MakeConnector(new Point(center.X + source.X, center.Y + source.Y), new Point(center.X + destination.X, center.Y + destination.Y), other));
			}
		}

        GeometryDrawing nodeDrawing = new GeometryDrawing();
        nodeDrawing.Pen = new Pen(Brushes.Black, 1);
        nodeDrawing.Brush = new SolidColorBrush(Colors.Red);

        GeometryGroup nodeGeometry = new GeometryGroup();

		// then draw all of the nodes
		foreach (Node node in mNodes) {
            Point destination = ScalePoint(node.Location, DrawScalingFactor);
            node.DisplayLocation = new Point(center.X + destination.X, center.Y + destination.Y);

			Size nodeSize = node.Size;
            node.DisplayRectangle = new Rect(
                node.DisplayLocation.X - (nodeSize.Width / 2), 
                node.DisplayLocation.Y - (nodeSize.Height / 2), 
                nodeSize.Width, 
                nodeSize.Height
            );

            nodes.Children.Add(node.MakeNode(node.DisplayRectangle));
		}

        drawing.Children.Add(connectors);
        drawing.Children.Add(nodes);

        return drawing;
	}

	/// <summary>
	/// Calculates the bearing angle from one point to another.
	/// </summary>
	/// <param name="start">The node that the angle is measured from.</param>
	/// <param name="end">The node that creates the angle.</param>
	/// <returns>The bearing angle, in degrees.</returns>
	private double GetBearingAngle(Point start, Point end) {
		Point half = new Point(start.X + ((end.X - start.X) / 2), start.Y + ((end.Y - start.Y) / 2));

		double diffX = (double)(half.X - start.X);
		double diffY = (double)(half.Y - start.Y);

		if (diffX == 0) diffX = 0.001;
		if (diffY == 0) diffY = 0.001;

		double angle;
		if (Math.Abs(diffX) > Math.Abs(diffY)) {
			angle = Math.Tanh(diffY / diffX) * (180.0 / Math.PI);
			if (((diffX < 0) && (diffY > 0)) || ((diffX < 0) && (diffY < 0))) angle += 180;
		}
		else {
			angle = Math.Tanh(diffX / diffY) * (180.0 / Math.PI);
			if (((diffY < 0) && (diffX > 0)) || ((diffY < 0) && (diffX < 0))) angle += 180;
			angle = (180 - (angle + 90));
		}

		return angle;
	}

	/// <summary>
	/// Determines the logical bounds of the diagram. This is used to center and scale the diagram when drawing.
	/// </summary>
	/// <returns>A System.Drawing.Rect that fits exactly around every node in the diagram.</returns>
	public Rect GetDiagramBounds() {
		double minX = Int32.MaxValue, minY = Int32.MaxValue;
        double maxX = Int32.MinValue, maxY = Int32.MinValue;
		foreach (Node node in mNodes) {
			if (node.X < minX) minX = node.X;
			if (node.X > maxX) maxX = node.X;
			if (node.Y < minY) minY = node.Y;
			if (node.Y > maxY) maxY = node.Y;
		}

		return new Rect(minX, minY, maxX - minX, maxY - minY);
	}

	/// <summary>
	/// Removes the specified node from the diagram. Any connected nodes will remain on the diagram.
	/// </summary>
	/// <param name="node">The node to remove from the diagram.</param>
	/// <returns>True if the node belonged to the diagram.</returns>
	public bool RemoveNode(Node node) {
		foreach (Node other in mNodes) {
			if ((other != node) && other.Connections.Contains(node)) other.Disconnect(node);
		}

        if (mNodes.Remove(node)) {
            node.Diagram = null;
            return true;
        }

        return false;
	}

	/// <summary>
	/// Applies a scaling factor to the specified point, used for zooming.
	/// </summary>
	/// <param name="point">The coordinates to scale.</param>
	/// <param name="scale">The scaling factor.</param>
	/// <returns>A System.Drawing.Point representing the scaled coordinates.</returns>
	private Point ScalePoint(Point point, double scale) {
		return new Point((int)((double)point.X * scale), (int)((double)point.Y * scale));
	}

    /// <summary>
    /// Returns the node at the specified location.
    /// </summary>
    /// <param name="location"></param>
    /// <returns></returns>
    public Node NodeAtLocation(Point location) {
        foreach (Node node in mNodes) {
            Rect bounds = new Rect(node.X - (node.Size.Width / 2), node.Y - (node.Size.Height / 2), node.Size.Width, node.Size.Height);
            if (bounds.Contains(location)) return node;
        }

        return null;
    }

	/// <summary>
	/// Private inner class used to track the node's position and velocity during simulation.
	/// </summary>
	private class NodeLayoutInfo {

		public Node Node;			// reference to the node in the simulation
		public Vector Velocity;		// the node's current velocity, expressed in vector form
		public Point NextPosition;	// the node's position after the next iteration

		/// <summary>
		/// Initialises a new instance of the Diagram.NodeLayoutInfo class, using the specified parameters.
		/// </summary>
		/// <param name="node"></param>
		/// <param name="velocity"></param>
		/// <param name="nextPosition"></param>
		public NodeLayoutInfo(Node node, Vector velocity, Point nextPosition) {
			Node = node;
			Velocity = velocity;
			NextPosition = nextPosition;
		}
	}
}