﻿// A Force-Directed Diagram Layout Algorithm
// Bradley Smith - 2010/07/01

using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Media;

/// <summary>
/// Represents a node that can be added to a diagram and connected to other nodes. 
/// Provides logic for drawing itself and its connections to other nodes.
/// </summary>
public abstract class Node {

	Diagram mDiagram;			// the parent diagram
	Point mLocation;			// node position, relative to the origin
	List<Node> mConnections;	// list of references to connected nodes (children)
    Rect mDisplayRectangle;
    Point mDisplayLocation;

	/// <summary>
	/// Gets or sets the position of the node.
	/// </summary>
	public Point Location {
		get { return mLocation; }
		set { mLocation = value; }
	}
	/// <summary>
	/// Gets a read-only collection representing the (child) nodes that this node is connected to.
	/// </summary>
	public IList<Node> Connections {
		get { return mConnections.AsReadOnly(); }
	}
	/// <summary>
	/// Gets or sets the diagram to which this node belongs.
	/// </summary>
	public Diagram Diagram {
		get { return mDiagram; }
		set {
			if (mDiagram == value) return;

			if ((mDiagram != null) && mDiagram.ContainsNode(this)) mDiagram.RemoveNode(this);
			mDiagram = value;
			if (mDiagram != null) mDiagram.AddNode(this);
		}
	}
	/// <summary>
	/// Gets or sets the X coordinate of the node, relative to the origin.
	/// </summary>
	public double X {
		get { return mLocation.X; }
		set { mLocation.X = value; }
	}
	/// <summary>
	/// Gets or sets the Y coordinate of the node, relative to the origin.
	/// </summary>
    public double Y {
		get { return mLocation.Y; }
		set { mLocation.Y = value; }
	}
	/// <summary>
	/// Gets the size of the node (for drawing purposes).
	/// </summary>
	public abstract Size Size {
		get;
	}
    /// <summary>
    /// Gets or sets the node's display rectangle.
    /// </summary>
    public Rect DisplayRectangle {
        get { return mDisplayRectangle; }
        set { mDisplayRectangle = value; }
    }
    /// <summary>
    /// Gets or sets the node's display location.
    /// </summary>
    public Point DisplayLocation {
        get { return mDisplayLocation; }
        set { mDisplayLocation = value; }
    }

	/// <summary>
	/// Initialises a new instance of the Node class.
	/// </summary>
	public Node() {
        mLocation = new Point();
		mDiagram = null;
		mConnections = new List<Node>();
	}

	/// <summary>
	/// Connects the specified child node to this node.
	/// </summary>
	/// <param name="child">The child node to add.</param>
	/// <returns>True if the node was connected to this node.</returns>
	public bool AddChild(Node child) {
		if (child == null) throw new ArgumentNullException("child");
		if ((child != this) && !this.mConnections.Contains(child)) {
			child.Diagram = this.Diagram;
			this.mConnections.Add(child);
			return true;
		}
		else {
			return false;
		}
	}

	/// <summary>
	/// Connects this node to the specified parent node.
	/// </summary>
	/// <param name="parent">The node to connect to this node.</param>
	/// <returns>True if the other node was connected to this node.</returns>
	public bool AddParent(Node parent) {
		if (parent == null) throw new ArgumentNullException("parent");
		return parent.AddChild(this);
	}

	/// <summary>
	/// Removes any connection between this node and the specified node.
	/// </summary>
	/// <param name="other">The other node whose connection is to be removed.</param>
	/// <returns>True if a connection existed.</returns>
	public bool Disconnect(Node other) {
		bool c = this.mConnections.Remove(other);
		bool p = other.mConnections.Remove(this);
		return c || p;
	}

	/// <summary>
	/// Generates a Drawing object representing the connector between two nodes.
	/// The source and destination coordinates (relative to the canvas) are also specified.
	/// </summary>
	/// <param name="from">Source coodinate.</param>
	/// <param name="to">Destination coordinate.</param>
	/// <param name="other">The other node.</param>
    public virtual Drawing MakeConnector(Point from, Point to, Node other) {
		return new GeometryDrawing(Brushes.Transparent, new Pen(Brushes.Black, 1), new LineGeometry(from, to));
	}

	/// <summary>
	/// Generates a Drawing object representing the node, within the specified bounds.
	/// </summary>
	/// <param name="graphics">GDI+ Graphics surface.</param>
	/// <param name="bounds">The bounds in which to draw the node.</param>
    public abstract Drawing MakeNode(Rect bounds);
}

/// <summary>
/// Provides an example implementation of the Node class. SpotNode is an 8x8 circle that is stroked and filled.
/// </summary>
public class SpotNode : Node {

	/// <summary>
	/// Gets the size of the spot.
	/// </summary>
	public override Size Size {
		get {
			return new Size(5, 5);
		}
	}
    /// <summary>
    /// Gets or sets the Brush to use when filling the spot.
    /// </summary>
    public Brush Fill { get; set; }
    /// <summary>
    /// Gets or sets the Pen to use when drawing the outline of the spot.
    /// </summary>
    public Pen Stroke { get; set; }

	/// <summary>
	/// Initalises a new instance of the SpotNode class using default values.
	/// </summary>
    public SpotNode() {
        Fill = Brushes.Red;
        Stroke = new Pen(Brushes.Black, 1);
    }

    /// <summary>
    /// Initalises a new instance of the SpotNode class using the specified fill color.
    /// </summary>
    /// <param name="color">The System.Windows.Media.Color to fill the spot with.</param>
    public SpotNode(Color color) {
        Fill = new SolidColorBrush(color);
        Stroke = new Pen(Brushes.Black, 1);
    }

	/// <summary>
	/// Generates a Drawing object representing the node.
	/// </summary>
	/// <param name="bounds">The bounds in which to draw the node.</param>
	public override Drawing MakeNode(Rect bounds) {
        return new GeometryDrawing(Fill, Stroke, new EllipseGeometry(bounds));
	}
}