﻿using System;
using System.Windows;

/// <summary>
/// Represents an immutable 2-dimensional vector, expressed in terms of either 
/// its X and Y components or magnitude and direction.
/// </summary>
public struct Vector {

    /// <summary>
    /// A vector whose magnitude is zero.
    /// </summary>
    public static readonly Vector Empty = new Vector();

    /// <summary>
    /// Gets the X (horizontal) component of the vector.
    /// </summary>
    public double X { get; private set; }
    /// <summary>
    /// Gets the Y (vertical) component of the vector.
    /// </summary>
    public double Y { get; private set; }
    /// <summary>
    /// Gets the magnitude of the vector.
    /// </summary>
    public double Magnitude {
        get {
            return Math.Sqrt(Math.Pow(X, 2) + Math.Pow(Y, 2));
        }
        private set {
            SetPolar(value, Direction);
        }
    }
    /// <summary>
    /// Gets the direction of the vector, in degrees.
    /// </summary>
    public double Direction {
        get {
            return (180.0 / Math.PI) * Math.Atan2(Y, X);
        }
        private set {
            SetPolar(Magnitude, value);
        }
    }

    /// <summary>
    /// Creates a vector from the specified Point structure.
    /// </summary>
    /// <param name="location"></param>
    public Vector(Point location)
        : this() {
        X = location.X;
        Y = location.Y;
    }

    /// <summary>
    /// Creates a vector from magnitude and direction values.
    /// </summary>
    /// <param name="magnitude"></param>
    /// <param name="direction"></param>
    public Vector(double magnitude, double direction) : this() {
        SetPolar(magnitude, direction);
    }

    /// <summary>
    /// Sets the X and Y components of the vector based on the specified magnitude and direction.
    /// </summary>
    /// <param name="magnitude"></param>
    /// <param name="direction"></param>
    private void SetPolar(double magnitude, double direction) {
        X = magnitude * Math.Cos((Math.PI / 180.0) * (double)direction);
        Y = magnitude * Math.Sin((Math.PI / 180.0) * (double)direction);
    }

    /// <summary>
    /// Adds two vectors and returns the resultant vector.
    /// </summary>
    /// <param name="a"></param>
    /// <param name="b"></param>
    /// <returns></returns>
    public static Vector Add(Vector a, Vector b) {
        Vector result = new Vector();
        result.X = a.X + b.X;
        result.Y = a.Y + b.Y;
        return result;
    }

    /// <summary>
    /// Multiplies a vector by a scalar value and returns the resultant vector.
    /// </summary>
    /// <param name="a"></param>
    /// <param name="b"></param>
    /// <returns></returns>
    public static Vector Multiply(Vector a, double b) {
        Vector result = new Vector();
        result.X = a.X * b;
        result.Y = a.Y * b;
        return result;
    }

    /// <summary>
    /// Returns the X and Y components of the vector as a Point structure.
    /// </summary>
    /// <returns></returns>
    public Point ToPoint() {
        return new Point(X, Y);
    }

    /// <summary>
    /// Negation operator.
    /// </summary>
    /// <param name="v"></param>
    /// <returns></returns>
    public static Vector operator -(Vector v) {
        Vector result = new Vector();
        result.X = -v.X;
        result.Y = -v.Y;
        return result;
    }

    /// <summary>
    /// Addition operator.
    /// </summary>
    /// <param name="a"></param>
    /// <param name="b"></param>
    /// <returns></returns>
    public static Vector operator +(Vector a, Vector b) {
        return Add(a, b);
    }

    /// <summary>
    /// Subtraction operator.
    /// </summary>
    /// <param name="a"></param>
    /// <param name="b"></param>
    /// <returns></returns>
    public static Vector operator -(Vector a, Vector b) {
        return Add(a, -b);
    }

    /// <summary>
    /// Multiplication operator (vector x scalar).
    /// </summary>
    /// <param name="a"></param>
    /// <param name="b"></param>
    /// <returns></returns>
    public static Vector operator *(Vector a, double b) {
        return Multiply(a, b);
    }

    /// <summary>
    /// Multiplication operator (scalar x vector).
    /// </summary>
    /// <param name="a"></param>
    /// <param name="b"></param>
    /// <returns></returns>
    public static Vector operator *(double a, Vector b) {
        return Multiply(b, a);
    }

    /// <summary>
    /// Returns a string representation of the vector, including its X and Y components, magnitude and direction.
    /// </summary>
    /// <returns></returns>
    public override string ToString() {
        return String.Format("({0:N3},{1:N3}) [{2:N3},{3:N1}°]", X, Y, Magnitude, Direction);
    }
}